/* Copyright (c) 2013 ACCESS CO., LTD. All rights reserved. */

var _ACS = _ACS || {};

_ACS.EPUBCFI = function() {

  function generateCFIForPosition($node, offset, classBlacklist, elementBlacklist, idBlacklist) {
    // If $node is inside a blacklisted node, we need to use the (outermost) blacklisted node
    // as the start node and adjust the offset accordingly.
    var $blacklistedAncestors = $node.parents().filter(function() {
      return EPUBcfi.CFIInstructions.isBlacklisted(this, classBlacklist, elementBlacklist, idBlacklist);
    });
    if ($blacklistedAncestors.length > 0) {
      var $virtualStartNode = $blacklistedAncestors.last();
      while (!$node.is($virtualStartNode)) {
        var $predecessors = EPUBcfi.CFIInstructions.allSiblingsBefore($node);
        offset += $predecessors.text().length;
        $node = $node.parent();
      }
    }

    return {
      elementSteps: EPUBcfi.Generator.createCFIElementSteps($node.parent(), 'html', classBlacklist, elementBlacklist, idBlacklist),
      textNodeStep: EPUBcfi.Generator.createCFITextNodeStep($node, offset, classBlacklist, elementBlacklist, idBlacklist)
    };
  }

  function generate(range, classBlacklist, elementBlacklist, idBlacklist) {
    var startCFI = generateCFIForPosition($acsJQuery(range.startContainer), range.startOffset, classBlacklist, elementBlacklist, idBlacklist);
    var endCFI = generateCFIForPosition($acsJQuery(range.endContainer), range.endOffset, classBlacklist, elementBlacklist, idBlacklist);

    var length = Math.min(startCFI.elementSteps.length, endCFI.elementSteps.length);
    var index = length;
    for (var i = 0; i < length; i++) {
      if (startCFI.elementSteps[i] !== endCFI.elementSteps[i]) {
        index = i === 0 ? 0 : startCFI.elementSteps[i - 1] === '/' ? i - 1 : i;
        break;
      }
    }
    var parentPath = startCFI.elementSteps.substring(0, index);
    var startSubPath = startCFI.elementSteps.substring(index) + startCFI.textNodeStep;
    var endSubPath = endCFI.elementSteps.substring(index) + endCFI.textNodeStep;
 
    return parentPath + ',' + startSubPath + ',' + endSubPath;
  }

  function generateFromXHTML(contentDocXhtml, packageDocXhtml, contentDocumentName, startTextNodeName, characterOffset) {
    var contentDoc = (new window.DOMParser).parseFromString(contentDocXhtml, 'application/xhtml+xml');
    var packageDoc = (new window.DOMParser).parseFromString(packageDocXhtml, 'text/xml');

    var startTextNode = $acsJQuery(startTextNodeName, contentDoc).contents()[1];
    var contentDocCFIComponent = EPUBcfi.Generator.generateCharacterOffsetCFIComponent(startTextNode, characterOffset); 
    var packageDocCFIComponent = EPUBcfi.Generator.generatePackageDocumentCFIComponent(contentDocumentName, packageDoc);
    var generatedCFI = EPUBcfi.Generator.generateCompleteCFI(packageDocCFIComponent, contentDocCFIComponent);
    return generatedCFI;
  }

  function getTargetElementId(contentDocXhtml, CFI) {
    var contentDoc = (new window.DOMParser).parseFromString(contentDocXhtml, 'application/xhtml+xml');

    var element = EPUBcfi.Interpreter.getTargetElement(CFI, contentDoc);
    return element.attr('id');
  }

  var getAllTextNodes = function($nodes, textNodes /*out*/) {
    textNodes = textNodes || [];
    $nodes.each(function () {
      // XXX: not supporting CDATA sections now
      if (this.nodeType == 3 /* Node.TEXT_NODE */) {
        textNodes.push(this);
      } else {
        getAllTextNodes($acsJQuery(this).contents(), textNodes);
      }
    });
    return textNodes;
  };

  var resolveTextNodeOffset = function($nodes, offset) {
    var textNodes = getAllTextNodes($nodes);

    for (var i in textNodes) {
      if (offset < 0) {
        break;
      }

      var textNode = textNodes[i];
      var text = textNode.nodeValue;
      if (offset <= text.length) {
        return { node: textNode, offset: offset };
      }
      offset -= text.length;
    }

    throw new RangeError("Text node offset out of range");
  };

  var splitRangeCfi = function(rangeCfi) {
    if (rangeCfi.slice(0, 8) !== "epubcfi(" || rangeCfi.slice(-1) !== ")") {
      return null;
    }
    var rangeCfiBody = rangeCfi.slice(8, -1);

    // We want to split the range CFI at commas, but not at commas that are
    // escaped or in brackets. We first split the range CFI with a special
    // regexp and re-join unnecessarily split parts. The splitting replaces
    // commas with "undefined" and leaves escapes and brackets intact. For
    // example, "a^,b,c[,]d" is split to ["a", "^,", "b", undefined, "c",
    // "[,]", "d"] and re-joined to ["a^,b", "c[,]d"].
    var rangeCfiParts = rangeCfiBody.split(/(\^.|\[(?:\^.|[^\^])*?\])|,/);
    var pathCfis = [], pathCfiParts = [];
    rangeCfiParts.push(undefined);
    for (var i in rangeCfiParts) {
      var part = rangeCfiParts[i];
      if (typeof part !== "undefined") {
        pathCfiParts.push(part);
      } else {
        pathCfis.push(pathCfiParts.join(""));
        pathCfiParts.length = 0;
      }
    }

    if (pathCfis.length !== 3) {
      return { start: rangeCfi, end: rangeCfi };
    }
    return {
      start: "epubcfi(" + pathCfis[0] + pathCfis[1] + ")",
      end: "epubcfi(" + pathCfis[0] + pathCfis[2] + ")"
    }
  };

  // This function is basically a mix of getTargetElement and getTextTerminusInfoWithPartialCFI.
  var getTextTerminusInfo = function(cfi, contentDocument, classBlacklist, elementBlacklist, idBlacklist) {
    var CFIAST = EPUBcfi.Parser.parse(cfi);

    // Skip to the first indirection step
    var indirectionStepNum = EPUBcfi.Interpreter.getFirstIndirectionStepNum(CFIAST);
    CFIAST.cfiString.localPath.steps[indirectionStepNum].type = "indexStep";

    // Interpret the rest of the steps
    var $nodes = EPUBcfi.Interpreter.interpretLocalPath(CFIAST.cfiString,
        indirectionStepNum, $acsJQuery("html", contentDocument),
        classBlacklist, elementBlacklist, idBlacklist);

    var offset = parseInt(CFIAST.cfiString.localPath.termStep.offsetValue);

    return resolveTextNodeOffset($nodes, offset);
  };

  var getTargetRange = function(rangeCfi, contentDocument, classBlacklist, elementBlacklist, idBlacklist) {
    if (!contentDocument) {
      contentDocument = document;
    }

    var cfiPair = splitRangeCfi(rangeCfi);
    if (!cfiPair) {
      return null;
    }

    var start, end;
    try {
      start = getTextTerminusInfo(cfiPair.start, contentDocument, classBlacklist, elementBlacklist, idBlacklist);
      end = getTextTerminusInfo(cfiPair.end, contentDocument, classBlacklist, elementBlacklist, idBlacklist);
    } catch (e) { // if CFI is invalid
      console.debug(e);
      return null;
    }

    var range = contentDocument.createRange();
    try {
      range.setStart(start.node, start.offset);
      range.setEnd(end.node, end.offset);
    } catch (e) {
      return null;
    }
    return range;
  }

  return {
    generate: generate,
    generateFromXHTML: generateFromXHTML,
    getTargetElementId: getTargetElementId,
    getTargetRange: getTargetRange
  };

}();
