/* Copyright(c) 2013 ACCESS CO., LTD. All rights reserved. */

if (typeof _ACS === "undefined") {
  window._ACS = {};
}

if (typeof _ACS.Utils === "undefined") {
  window._ACS.Utils = {};
}

window._ACS.Utils.Rect = function(left, top, width, height) {
  this.left = left;
  this.top = top;
  this.width = width;
  this.height = height;

  this.right = function() {
    return this.left + this.width;
  }

  this.bottom = function() {
    return this.top + this.height;
  }
}

window._ACS.Bookmark = (function() {
  "use strict";
  var SAMPLENG_COUNT = 5;
  function getACSPosition(visibleRect, options) {
    if (options === undefined) {
      options = {};
    }
    var doc = document;
    if (options.document != undefined) {
      doc = options.document;
    }
    var isVerticalDocument = false;
    if (options.isVerticalDocument != undefined) {
      isVerticalDocument = options.isVerticalDocument;
    }
    if (typeof doc.caretRangeFromPoint === "undefined") {
      return null;
    }

    if (options.debugPointMode) {
      var startRange = doc.caretRangeFromPoint(visibleRect.left, visibleRect.top);
      return JSON.stringify(getACSPositionByRange(doc, startRange));
    }

    var samplingPointX = visibleRect.left;
    var samplingPointY = visibleRect.top;
    var bodyMarginTop = parseInt($acsJQuery(doc.body).css("margin-top"));
    var bodyMarginLeft = parseInt($acsJQuery(doc.body).css("margin-left"));
    var bodyMarginBottom = parseInt($acsJQuery(doc.body).css("margin-bottom"));
    var bodyMarginRight = parseInt($acsJQuery(doc.body).css("margin-right"));
    var validWidth = visibleRect.width - bodyMarginLeft - bodyMarginRight - 1;
    var validHeight = visibleRect.height - bodyMarginTop - bodyMarginBottom - 1;

    samplingPointY += bodyMarginTop;
    if (isVerticalDocument) {
      samplingPointX = visibleRect.right() - bodyMarginRight - 1;
      // for iOS
      // var offset = $acsJQuery(doc).width() - this.lastViewPortSize.width;
      // samplingPointX += offset;
    } else {
      samplingPointX += bodyMarginLeft;
    }

    var startRange = doc.caretRangeFromPoint(samplingPointX, samplingPointY);

    if (canUseRangeForBookmark(startRange, true)) {
      return JSON.stringify(getACSPositionByRange(doc, startRange));
    }

    var horizontalAdvance = validWidth / (SAMPLENG_COUNT - 1);
    if (isVerticalDocument) {
      horizontalAdvance = -horizontalAdvance;
    }
    var verticalAdvance = validHeight / (SAMPLENG_COUNT - 1);
    for (var count = 1; count < SAMPLENG_COUNT; count++) {
      samplingPointX += horizontalAdvance;
      samplingPointY += verticalAdvance;
      var range = doc.caretRangeFromPoint(samplingPointX, samplingPointY);
      if (canUseRangeForBookmark(range, false)) {
        range.setStart(range.startContainer, 0);
        return JSON.stringify(getACSPositionByRange(doc, range));
      }
    }

    return JSON.stringify(getACSPositionByRange(doc, startRange));
  }

  function canUseRangeForBookmark(range, isTop) {
    if (!range) {
      return false;
    }
    var startNode = range.startContainer;
    var startOffset = range.startOffset;
    if (startNode.nodeType === 1 /* Node.ELEMENT_NODE */) {
      if (startNode.nodeName.toLowerCase() === "img" ||
        startNode.nodeName.toLowerCase() === "image") {
        return true;
      }
      if (startNode.childNodes.length === 0) {
        // startNode is empty node.
        return false;
      }
      if (startNode.childNodes.length >= startOffset) {
        // end of startNode.
        // Do not use this position.
        return false;
      }
      var node = startNode.childNodes[startOffset];
      if (node.nodeType === 1 /* Node.ELEMENT_NODE */ &&
        (node.nodeName.toLowerCase() === "img" || node.nodeName.toLowerCase() === "image")) {
        return true;
      }
      if (node.nodeType === 3 /* Node.TEXT_NODE */) {
        return true;
      }
    } else if (startNode.nodeType === 3 /* Node.TEXT_NODE */) {
      if (!isTop) {
        return true;
      }

      if (startOffset === 0) {
        // beginning of text node.
        return true;
      }
      var text = startNode.nodeValue;
      if (startOffset === text.length) {
        // end of text node.
        // It is the end of the previous page.
        return false;
      }

      var substr = text.substr(startOffset - 1, 3);
      if (substr.match(/^\s|\s$/)) {
        // Do not use when there is a space before and after.
        return false;
      }
      return true;
    }
    return false;
  }

  function highlightClassName() {
    if (_ACS.Selection.HIGHLIGHT_CLASS_NAME) {
      return _ACS.Selection.HIGHLIGHT_CLASS_NAME;
    }
    return '-acs-highlight';
  }

  function isHighlightNode(node) {
    return $acsJQuery(node).hasClass(highlightClassName());
  }

  function getACSPositionByRange(doc, range) {
    if (!range) {
      return null;
    }
    var b = getACSPositionOrNode(doc, range, -1);
    if (b === 0) {
      return null;
    }
    return {nodeIndex: b.nodeIndex, offset: b.offset};
  }

  function getACSPositionOrNode(doc, range, index) {
    if (!range && index < 0) {
      return {nodeIndex: 0, offset: 0, node: null};
    }
    var node = range ? range.startContainer : null;
    var offset = 0;
    var startOffset = range ? range.startOffset : 0;
    var walker = doc.createTreeWalker(doc.documentElement, NodeFilter.SHOW_ALL);

    var currentNode = walker.currentNode;
    var nodeIndex = -1;
    var prevNode = walker.currentNode;
    var STATUS = {none: 0, inHighlight: 1, inHighlightWithCombinedTextNode: 2}
    var status = STATUS.none;

    while (currentNode) {
      switch (status) {
        case STATUS.none:
          if (isHighlightNode(currentNode)) {
            status = STATUS.inHighlight;
            break;
          }
          status = STATUS.none;
          prevNode = currentNode;
          nodeIndex++;
          offset = 0;
          break;
        case STATUS.inHighlight:
        case STATUS.inHighlightWithCombinedTextNode:
          if (isText(currentNode)) {
            if (!isText(prevNode)) {
              nodeIndex++;
              offset = 0;
            } else {
              offset += prevNode.textContent.length;
              status = STATUS.inHighlightWithCombinedTextNode; 
            }
            prevNode = currentNode;
          }
          if (!withinHighlightedNode(currentNode)) {
            status = STATUS.none;
          }
          break;
        default:
          throw "Unreachable";
      }
      if ((node && currentNode === node) ||
          (!node && nodeIndex === index)) {
        if (status === STATUS.inHighlightWithCombinedTextNode) {
          offset += currentNode.textContent.length;
        }
        return {nodeIndex: nodeIndex, offset: startOffset + offset, node: currentNode};
      }
      currentNode = walker.nextNode();
    }
    return {nodeIndex: 0, offset: 0, node: null};
  }

  function withinHighlightedNode(node) {
    if (isHighlightNode(node)) {
      return true;
    }
    var $highlightedAncestors = $acsJQuery(node).parents().filter(function() {
      return isHighlightNode($acsJQuery(this));
    });
    return $highlightedAncestors.length > 0;
  }

  function nextTextNode(root, node) {
    var doc = node.ownerDocument;
    var walker = doc.createTreeWalker(root, NodeFilter.SHOW_ALL);
    var currentNode = walker.currentNode = node;
    while (currentNode = walker.nextNode()) {
      if (isText(currentNode)) {
        return currentNode;
      }
    }
    return null;
  }

  function isText(node) {
    return node.nodeType === 3 /* Node.TEXT_NODE */;
  }

  function getRectByNodeAndOffset(node, offset, options) {
    options = options || {};
    var doc = options.document || document;
    if (!node) {
      return null
    }
    var parent = node.parentNode;
    var range = doc.createRange();
    var textnode = isText(node) ? node : nextTextNode(parent, node)
    while (textnode && offset > textnode.textContent.length) {
      offset -= textnode.textContent.length;
      textnode = nextTextNode(parent, textnode);
    }
    // for safety
    if (!textnode) {
      textnode = node;
      if (isText(textnode)) {
        offset = textnode.textContent.length - 1;
      } else {
        offset = 0;
      }
    } else if (offset >= textnode.textContent.length &&
        offset > 0) {
      offset = textnode.textContent.length - 1;
    }

    // for more safety: find a text node containing a non-space character.
    while (!isText(textnode) || !/\S/.test(textnode.nodeValue)) {
      var next = nextTextNode(doc.documentElement, textnode);
      if (!next) {
        break;
      }
      textnode = next;
      offset = 0;
    }
    // and the offset should point to a non-space character.
    if (isText(textnode) && !/\S/.test(textnode.nodeValue.charAt(offset))) {
      while (offset > 0) {
        var tail = textnode.nodeValue.substring(offset);
        var index = tail.search(/\S/);
        if (index >= 0) {
          offset += index;
          break;
        }
        offset--;
      }
    }

    range.setStart(textnode, offset);
    // TODO: setEnd could throw an error when the offset is grater than
    //       length of text.
    range.setEnd(textnode, offset + 1);
    return range.getBoundingClientRect();
  }

  function getBookmarkPoint(nodeIndex, offset, options) {
    options = options || {};
    var doc = options.document || document;
    var win = options.window || window;
    var b = getACSPositionOrNode(doc, null, nodeIndex);
    var rect = b.node ? getRectByNodeAndOffset(b.node, offset, options) : null;
    if (!rect) {
      return null;
    }
    return JSON.stringify({left: rect.left + win.scrollX, top: rect.top + win.scrollY});
  }

  function getNodeForCharCount(bodyNode, charCount) {
    var currentCharCount = 0;
    var walker = bodyNode.ownerDocument.createTreeWalker(
        bodyNode, NodeFilter.SHOW_ELEMENT | NodeFilter.SHOW_TEXT,
        { acceptNode: function(node) {
            switch (node.nodeName) {
                case "rp":
                case "rt":
                case "script":
                case "style":
                    return NodeFilter.FILTER_REJECT;
                default:
                    return NodeFilter.FILTER_ACCEPT;
            }
        } });

    var node;
    var lastTextNode = bodyNode;
    while (node = walker.nextNode()) {
      if (node.nodeName.toLowerCase() === 'br') {
        currentCharCount++;
        continue;
      } else if (node.nodeType !== Node.TEXT_NODE) {
        continue;
      }
      lastTextNode = node;
      var text = node.nodeValue;
      text = text.replace(/[ \t\n\r]+/g, " ");
      var length = text.length;
      var offset = charCount - currentCharCount;
      if (offset <= length) {
        return { node: node, offset: offset };
      }
      currentCharCount += length;
    }

    // Node not found. Fallback to the last node.
    if (lastTextNode) {
      return { node: lastTextNode, offset: lastTextNode.nodeValue.length };
    }
    return { node: bodyNode, offset: bodyNode.childNodes.length };
  }

  function getCharIndexPoint(stringCharIndex) {
    var contentDocument = document;
    var win = window;
    var result = getNodeForCharCount(contentDocument.body, stringCharIndex);
    var rect = result.node ? getRectByNodeAndOffset(result.node, result.offset, null) : null;
    if (!rect) {
      return JSON.stringify({left: -1, top: -1});
    }
    return JSON.stringify({left: rect.left + win.scrollX, top: rect.top + win.scrollY});
  }

  return {
      getACSPosition: getACSPosition,
      getBookmarkPoint: getBookmarkPoint,
      getCharIndexPoint: getCharIndexPoint,
  };
})();
