/* Copyright (c) 2013 ACCESS CO., LTD. All rights reserved. */

var _ACS = _ACS || {};

_ACS.Selection = function() {

  var HIGHLIGHT_CLASS_NAME = '-acs-highlight';
  var HIGHLIGHT_TYPE_NORMAL = 'normal';
  var HIGHLIGHT_TYPE_EXTRA = 'extra';

  /** Mapping from highlight IDs to highlight data. */
  var highlights = {};

  function getCurrentSelectionRange(doc) {
    if (doc == undefined) doc = document;
    var selection = doc.getSelection();
    if (selection.rangeCount === 0) {
      return;
    }
    return selection.getRangeAt(0);
  }

  function wrapTextNodes(textNodes, wrapper) {
    return $acsJQuery(textNodes)
      .filter(function(index) {
        return /\S/.test(this.nodeValue);
      })
      .wrap(wrapper);
  }

  function $highlightWrapper(id, color) {
    return $acsJQuery('<span/>')
    .addClass(HIGHLIGHT_CLASS_NAME)
    .css({ "background-color": color, "-epub-text-combine": "inherit" })
    .attr('data-acs-highlight-id', id)
    .click(function() {
      var highlight = highlights[this.dataset.acsHighlightId];
      if (highlight) {
        _ACS.callback('highlight-tapped:' + JSON.stringify(highlight));
      }
    });
  }

  function highlightCurrentSelection(id, color, doc) {
    var range = Range.sniff(getCurrentSelectionRange(doc));
    if (!range) {
      return;
    }
    wrapTextNodes(range.normalize().textNodes(), $highlightWrapper(id, color));
  }

  function clearExtraHighlights() {
    for (var i in highlights) {
      if (highlights[i].type === HIGHLIGHT_TYPE_EXTRA) {
        clearHighlight(highlights[i]);
      }
    }
  }
  
  function clearHighlight(args) {
    var id = args.id;
    $acsJQuery('[data-acs-highlight-id="' + _ACS.escapeJQuerySelector(id) + '"]')
      .contents()
      .unwrap() // remove highlighting span elements
      .parent().each(function() {
        this.normalize(); // join split text nodes
      });
  }

  function joinText(textNodes) {
    return $acsJQuery(textNodes)
      .filter(function() { // remove nodes contained in rt or rp
        return $acsJQuery(this).closest("rt").length === 0 &&
               $acsJQuery(this).closest("rp").length === 0;
      })
      .map(function() {
        return this.nodeValue;
      })
      .get()
      .join("");
  }

  function getSelectedText(doc) {
    var range = getCurrentSelectionRange(doc);
    if (!range || range.collapsed) {
      return null;
    }
    range = Range.sniff(range);
    if (!range) {
      return "";
    }
    range = range.normalize(); // may split text nodes

    var text = joinText(range.textNodes());

    // Restore to the original DOM tree
    if (range.start && range.start.parentNode) {
      range.start.parentNode.normalize();
    }
    if (range.end && range.end.parentNode) {
      range.end.parentNode.normalize();
    }

    return text;
  }

  function addHighlightFromCurrentSelection(args) {
    var id = args.id;
    var color = args.color;
    var packageDocumentCfiComponent = args.cfi;
    var doc = args.document;

    var range = getCurrentSelectionRange(doc);
    if (!range || range.collapsed) {
      return null;
    }

    var contentDocCfiComponent = _ACS.EPUBCFI.generate(range, [HIGHLIGHT_CLASS_NAME]);
    var cfi = "epubcfi(" + packageDocumentCfiComponent + contentDocCfiComponent + ")";

    return addHighlightFromRange(id, color, cfi, range, HIGHLIGHT_TYPE_NORMAL);
  }

  function addExtraHighlightFromEPUBCFI(args) {
    var id = args.id;
    var color = args.color;
    var cfi = args.epubcfi;

    var range = _ACS.EPUBCFI.getTargetRange(cfi, document, [HIGHLIGHT_CLASS_NAME]);
    if (!range || range.collapsed) {
        return null;
    }

    return addHighlightFromRange(id, color, cfi, range, HIGHLIGHT_TYPE_EXTRA);
  }

  function addHighlightFromRange(id, color, cfi, range, type) {
    range = Range.sniff(range);
    if (!range) {
      return null;
    }
    var text = highlightTextFromRange(id, color, range, type);

    var highlight = {
      format: 1,
      id: id,
      type: type,
      text: text,
      color: color,
      ranges: [ cfi ]
    };
    highlights[id] = highlight;
    return JSON.stringify(highlight);
  }
  
  function highlightTextFromRange(id, color, range, type) {
    var textNodes = range.normalize().textNodes();
    var text = joinText(textNodes);
    wrapTextNodes(textNodes, $highlightWrapper(id, color));
    return text;
  }
  
  function addHighlight(highlight, doc) {
    if (highlight instanceof Array) {
      for (var i in highlight) {
        addHighlight(highlight[i], doc);
      }
      return;
    }

    if (!doc) {
      doc = document;
    }
    
    if (!highlight || !highlight.id || !highlight.color || !highlight.ranges) {
      return;
    }

    var wrapper = $highlightWrapper(highlight.id, highlight.color);
    for (var i in highlight.ranges) {
      var cfi = highlight.ranges[i];
      var range = _ACS.EPUBCFI.getTargetRange(cfi, doc, [HIGHLIGHT_CLASS_NAME]);
      if (range && !range.collapsed) {
        range = Range.sniff(range);
        if (range) {
          wrapTextNodes(range.normalize().textNodes(), wrapper);
        }
      }
    }

    highlights[highlight.id] = highlight;
  }

  function getCFIOfCurrentSelectionRange(doc) {
    var range = getCurrentSelectionRange(doc);
    if (!range || range.collapsed) {
      return null;
    }
    range = Range.sniff(range);
    if (!range) {
      return "";
    }

    return _ACS.EPUBCFI.generate(range, [HIGHLIGHT_CLASS_NAME]);
  }

  function getPositionFromCFI(args, doc) {
    if (doc === undefined) doc = document;
    var cfi = args.epubcfi;
    var range = _ACS.EPUBCFI.getTargetRange(cfi, doc, [HIGHLIGHT_CLASS_NAME]);
    var rect = range.getBoundingClientRect();

    return JSON.stringify({left: rect.left + window.scrollX, top: rect.top + window.scrollY,
                 right: rect.right + window.scrollX, bottom: rect.bottom + window.scrollY});
  }

  function resolveAbsoluteURL(relativeUrl) {
	var a = document.createElementNS('http://www.w3.org/1999/xhtml', 'a');
	a.href = relativeUrl;
	return a.href;
  }

  function getHref(element) {
    if (element.href instanceof SVGAnimatedString) {
      return resolveAbsoluteURL(element.href.baseVal);
    }
    return element.href;
  }

  var downTime, eventTime;
  var touchstartX, touchstartY;
  var threshold = 20;
  var doubletapTimeout = 300;
  var doubletapTimer = null;
  var timerIsOn = false;

  var tappedAnchor;
  var touchAnchorStartTime;
  var touchAnchorStartX, touchAnchorStartY;
  var anchorTappedTimeout = 300;
  var anchorTapConfirmed = false;
  var anchorDoubleTapTimer = null;
  var timerAnchorIsOn = false;
  var iframePos = 0;

  function setUpTouchEventListeners(args) {
    downTime = 0;
    eventTime = 0;
    touchstartX = 0;
    touchstartY = 0;
    doubletapTimer = null;
    timerIsOn = false;

    tappedAnchor = null;
    touchAnchorStartTime = 0;
    touchAnchorStartX = 0;
    touchAnchorStartY = 0;
    anchorTapConfirmed = false;
    anchorDoubleTapTimer = null;
    timerAnchorIsOn = false;

    if (args.enabled === true) {
        setAnchorListeners();
        addCustomTouchEventListeners();
    } else {
        removeAnchorListeners();
        removeCustomTouchEventListeners();
    }
  }

  function linkTappedHandler(e) {
    _ACS.callback('link-tapped:' + getHref(this));
  }

  function setAnchorListeners() {
    var anchors = document.getElementsByTagName('a');
    for(var i = 0, len = anchors.length; i < len; i++) {
      var anchor = anchors[i];
      anchor.addEventListener('touchstart', anchorTapStart, false);
      anchor.addEventListener('touchend', anchorTapEnd, false);
    }
  }

  function removeAnchorListeners() {
    var anchors = document.getElementsByTagName('a');
    for(var i = 0, len = anchors.length; i < len; i++) {
      var anchor = anchors[i];
      anchor.removeEventListener('touchstart', anchorTapStart, false);
      anchor.removeEventListener('touchend', anchorTapEnd, false);
    }
  }

  function anchorTapStart(e) {
    tappedAnchor = this;
    anchorTapConfirmed = false;
    touchAnchorStartTime = event.timeStamp;
    touchAnchorStartX = event.pageX;
    touchAnchorStartY = event.pageY;
  }

  function anchorTapEnd() {
    var touchEndTime = event.timeStamp;
    var href = getHref(tappedAnchor);
    if((touchEndTime - touchAnchorStartTime) < anchorTappedTimeout) {
      if (Math.abs(touchAnchorStartX - event.pageX) < threshold
          && Math.abs(touchAnchorStartY - event.pageY) < threshold) {
         anchorTapConfirmed = true;
         anchorEndTapHandler(event);
      }
    }
  }

  function anchorEventHandler() {
    if (anchorTapConfirmed && typeof tappedAnchor !== undefined ) {
      var href = getHref(tappedAnchor);
      _ACS.callback('link-tapped:' + href);
      tappedAnchor = undefined;
    }
  }

  function anchorEndTapHandler(event) {
    if (!timerAnchorIsOn) {
      timerAnchorIsOn = true;
      anchorDoubleTapTimer = setTimeout(function () {
          anchorEventHandler();
          clearTimeout(anchorDoubleTapTimer);
          timerAnchorIsOn = false;
      }, doubletapTimeout);
    } else {
      clearTimeout(anchorDoubleTapTimer);
      timerAnchorIsOn = false;
      doubleTapHandler(event);
    }
  }

  function addCustomTouchEventListeners() {
    document.addEventListener('touchstart', singleTapHandler, false);
    document.addEventListener('touchend', tapHandler, false);
    var iframe = $acsJQuery("iframe");
    if (iframe != null) {
        iframePos = iframe.position();
        iframe.contents().on({
            touchstart: singleTapHandler,
            touchend: tapHandler,
        });
    }
  }

  function removeCustomTouchEventListeners() {
    document.removeEventListener('touchstart', singleTapHandler, false);
    document.removeEventListener('touchend', tapHandler, false);
  }

  function tapHandler(e) {
    if (!timerIsOn) {
      timerIsOn = true;
      doubletapTimer = setTimeout(function () {
          singleTapHandler(e);
          clearTimeout(doubletapTimer);
          timerIsOn = false;
      }, doubletapTimeout);
    } else {
      clearTimeout(doubletapTimer);
      timerIsOn = false;
      doubleTapHandler(e)
    }
  }

  function isChildNodeOfLinkElement(target) {
    return $acsJQuery(target).parents().filter(function() {
      return this.localName === "a";
    }).length > 0;
  }

  function singleTapHandler(e) {
    if (e.type === "touchstart") {
      downTime = e.timeStamp;
      touchstartX = e.pageX;
      touchstartY = e.pageY;
      if (touchstartX == null || touchstartY == null) {
          var touch = e.originalEvent.touches[0];
          touchstartX = touch.pageX;
          touchstartY = touch.pageY;
      }
    } else if (e.type === "touchend") {
      if (e.target.localName != "a" && e.target.className != HIGHLIGHT_CLASS_NAME &&
          !isChildNodeOfLinkElement(e.target)) {
         eventTime = e.timeStamp;
         var posX = e.pageX;
         var posY = e.pageY;
         if (posX == null || posY == null) {
            var touch = e.originalEvent.changedTouches[0];
            posX = touch.pageX;
            posY = touch.pageY;
         }
         if (Math.abs(touchstartX - posX) < threshold
             && Math.abs(touchstartY - posY) < threshold)
             hasEventsOnPoint(e);
      }
    }
  }

  function doubleTapHandler(e) {
    eventTime = e.timeStamp;
    var posX = e.pageX;
    var posY = e.pageY;
    _ACS.callback('body-double-tapped:' + JSON.stringify({x:posX, y:posY,
          eventTime:eventTime, downTime:downTime}));
  }

  function hasEventsOnPoint(e) {
    var hasEventElements = [/*"a", "area", */"button", "canvas", "input", "select", "textarea"];
    var element = e.target;
    var eventName;
    var elementName = element.nodeName.toLowerCase();
    var isEventElement;
    var posX = e.pageX;
    var posY = e.pageY;
    if (posX == null || posY == null) {
        var touch = e.originalEvent.changedTouches[0];
        posX = touch.pageX + iframePos.left;
        posY = touch.pageY + iframePos.top;
    }
    for (var i = 0; i < hasEventElements.length; i++) {
        isEventElement = Boolean(elementName === hasEventElements[i] ||
            $acsJQuery(element).parents(hasEventElements[i]).length > 0);
        if (element.onclick || $acsJQuery._data(element, "events")) {
            console.log("JavaScript element "+element.tagName+" events");
            isEventElement = true;
            elementName = "OnClick";
        } else if (element.tagName != "html" && element.tagName != "body" && element.addEventListener) {
            console.log("JavaScript element "+element.tagName+" addEventListener");
            isEventElement = true;
            elementName = "OnClick";
        } // TODO element.addEventListener いつもTrueになって、ページめくれない可能性がある。チケット #50251

        if (isEventElement) {
            _ACS.callback('body-tapped:' + JSON.stringify({x:posX, y:posY,
                eventTime:eventTime, downTime:downTime, elementName:elementName}));
            break;
        }
    }
    if (!isEventElement) {
        _ACS.callback('body-tapped:' + JSON.stringify({x:posX, y:posY,
                eventTime:eventTime, downTime:downTime}));
    }
  }


  return {
    addHighlight: addHighlight,
    addHighlightFromCurrentSelection: addHighlightFromCurrentSelection,
    addExtraHighlightFromEPUBCFI: addExtraHighlightFromEPUBCFI,
    clearHighlight: clearHighlight,
    clearExtraHighlights: clearExtraHighlights,
    getCFIOfCurrentSelectionRange: getCFIOfCurrentSelectionRange,
    getSelectedText: getSelectedText,
    highlightCurrentSelection: highlightCurrentSelection,
    getPositionFromCFI: getPositionFromCFI,
    setUpTouchEventListeners: setUpTouchEventListeners,
    addCustomTouchEventListeners: addCustomTouchEventListeners,
    removeCustomTouchEventListeners: removeCustomTouchEventListeners
  };

}();
